import type { LEDGERS } from '../../config/ledger.js';
import type {
  IndyVdrDidCreateOptions,
  IndyVdrDidCreateResult,
} from '@aries-framework/indy-vdr';
import type {
  EventDidsRegisterIndyFromSeed,
  EventDidsRegisterIndyFromSeedInput,
  EventDidsResolve,
  EventDidsResolveInput,
} from '@ocm/shared';

import {
  DidDocumentService,
  Hasher,
  KeyType,
  TypedArrayEncoder,
} from '@aries-framework/core';
import { Injectable } from '@nestjs/common';
import { ConfigService } from '@nestjs/config';

import { registerPublicDids } from '../ledger/register.js';
import { WithTenantService } from '../withTenantService.js';

@Injectable()
export class DidsService {
  public constructor(
    private withTenantService: WithTenantService,
    private configService: ConfigService,
  ) {}

  public async resolve({
    did,
    tenantId,
  }: EventDidsResolveInput): Promise<EventDidsResolve['data']> {
    return this.withTenantService.invoke(tenantId, async (t) => {
      const {
        didDocument,
        didResolutionMetadata: { message, error },
      } = await t.dids.resolve(did);

      if (!didDocument) {
        throw new Error(
          `Could not resolve did: '${did}'. Error: ${
            error ?? 'None'
          } Message: ${message ?? 'None'}`,
        );
      }

      return didDocument;
    });
  }

  public async registerDidIndyFromSeed({
    tenantId,
    seed,
    services,
  }: EventDidsRegisterIndyFromSeedInput): Promise<
    EventDidsRegisterIndyFromSeed['data']
  > {
    const dids: Array<string> = [];

    const ledgerIds = this.configService.get('agent.ledgerIds') as Array<
      keyof typeof LEDGERS
    >;

    const publicDidSeed = this.configService.get(
      'agent.publicDidSeed',
    ) as string;

    const publicDids = await registerPublicDids({
      ledgerIds,
      seed: publicDidSeed,
    });

    const { publicKey, publicKeyBase58 } = await this.withTenantService.invoke(
      tenantId,
      async (t) =>
        t.wallet.createKey({
          privateKey: TypedArrayEncoder.fromString(seed),
          keyType: KeyType.Ed25519,
        }),
    );

    await this.withTenantService.invoke(tenantId, async (t) =>
      t.wallet.createKey({
        privateKey: TypedArrayEncoder.fromString(publicDidSeed),
        keyType: KeyType.Ed25519,
      }),
    );

    const buffer = Hasher.hash(publicKey, 'sha2-256');
    const id = TypedArrayEncoder.toBase58(buffer.slice(0, 16));
    const verkey = publicKeyBase58;

    for (const publicDid of publicDids) {
      const did = `did:indy:${publicDid.namespace}:${id}`;
      const didDocumentServices: Array<DidDocumentService> | undefined =
        services?.map(
          (s) =>
            new DidDocumentService({
              id: `${did}#${s.identifier}`,
              type: s.type,
              serviceEndpoint: s.url,
            }),
        );


      await this.withTenantService.invoke(tenantId, async (t) => {
        const result = (await t.dids.create<IndyVdrDidCreateOptions>({
          did,
          options: {
            verkey,
            endorserMode: 'internal',
            endorserDid: publicDid.did,
            services: didDocumentServices,
            useEndpointAttrib: true,
          },
        })) as IndyVdrDidCreateResult;

        if (result.didState.state !== 'finished') {
          throw Error(
            `An error occurred while trying to register the did: '${did}'. Result: ${JSON.stringify(result)}`,
          );
        }

        dids.push(result.didState.did);
      });
    }

    return dids;
  }
}
